#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/eventfd.h>
#include <sys/wait.h>
#include <signal.h>
#include "nfqueue.h"
#include "processor.h"

#define exit(a) do { fflush(stdout); fflush(stderr); _exit(a); } while(0)

static int sfd;
static int term=0;

static void ch(int unused)
{
	int status;
	uint64_t dummy;

	while(waitpid(-1,&status,WNOHANG)>0)
	{
		dummy=1;
		dummy=write(sfd,&dummy,sizeof(dummy));
	}
}

static void dh(int unused)
{
	term=1;
}

static void usage(void)
{
	fprintf(stderr,"Usage: netctld [-f] [-d] [-h] [-H] [-c cfgfile] "
		"[-w maxwait] [-x script]\n"
			"       netctld -t\n"
			"       netctld -s device\n"
			"       netctld -4 device\n"
			"       netctld -6 device\n"
		"-f          don't daemonize (default if -d specified)\n"
		"-d          debug mode, don't call script, print events\n"
		"-h          hack for Samsung erroneus DELIF events\n"
		"-H          hack for Samsung complete wlan device removal\n"
		"-c cfgfile  configuration file (default: /etc/netctld.conf)\n"
		"-w maxwait  seconds to wait for script (1-99, default: 10)\n"
		"-x script   script to execute (default: /etc/netctld.sh)\n"
		"-t          request daemon termination\n"
		"-s device   print up/down status of device\n"
		"-4 device   print IPv4 adresses of device\n"
		"-6 device   print IPv6 adresses of device\n");
}

int main(int argc,char *argv[])
{
	int c;
	int err=1;
	int efd;
	int queue[2];
	int nfh[3];
	int pfh[2];
	pthread_t nfqthr;
	pthread_t prcthr;
	pthread_t unxthr;
	uint64_t dummy=1;
	sigset_t s;
	char *cfgfile=NULL;
	char *script=NULL;
	int maxwait=-1;
	int trace=0;
	int fg=0;
	int trq=0;
	int hack=0;
	char *srq=NULL;
	char *v4rq=NULL;
	char *v6rq=NULL;

	sigfillset(&s);
	sigprocmask(SIG_BLOCK,&s,NULL);
	signal(SIGHUP,dh);
	signal(SIGINT,dh);
	signal(SIGQUIT,dh);
	signal(SIGTERM,dh);
	signal(SIGCHLD,ch);

	while((c=getopt(argc,argv,"c:x:w:dfhHts:4:6:"))!=-1)switch(c)
	{
	case 'h':
		if(hack&1)goto err;
		hack|=1;
		break;

	case 'H':
		if(hack&2)goto err;
		hack|=2;
		break;

	case 'c':
		if(cfgfile)goto err;
		cfgfile=optarg;
		break;

	case 'x':
		if(script)goto err;
		script=optarg;
		break;

	case 'w':
		if(maxwait!=-1)goto err;
		maxwait=atoi(optarg);
		if(maxwait<1||maxwait>99)goto err;
		break;

	case 'd':
		if(trace)goto err;
		trace=1;
		break;

	case 'f':
		if(fg)goto err;
		fg=1;
		break;

	case 't':
		if(trq)goto err;
		trq=1;
		break;

	case 's':
		if(srq)goto err;
		srq=optarg;
		break;

	case '4':
		if(v4rq)goto err;
		v4rq=optarg;
		break;

	case '6':
		if(v6rq)goto err;
		v6rq=optarg;
		break;;
		
	default:fprintf(stderr,"\n");
err:		usage();
		goto err1;
	}

	switch(trq+(srq?1:0)+(v4rq?1:0)+(v6rq?1:0))
	{
	case 1:	if(maxwait!=-1||script||cfgfile||trace||fg)
		{
			usage();
			goto err1;
		}

	case 0:	break;

	default:usage();
		goto err1;
	}

	if(trq)exit(termreq()?1:0);
	if(srq)exit(statreq(srq)?1:0);
	if(v4rq)exit(v4req(v4rq)?1:0);
	if(v6rq)exit(v6req(v6rq)?1:0);

	if(!cfgfile)cfgfile="/etc/netctld.conf";
	if(!script)script="/etc/netctld.sh";
	if(maxwait==-1)maxwait=10;
	if(trace)fg=1;

	if((efd=eventfd(0,EFD_CLOEXEC))==-1)
	{
		perror("eventfd");
		goto err1;
	}

	if((sfd=eventfd(0,EFD_CLOEXEC|EFD_NONBLOCK))==-1)
	{
		perror("eventfd");
		goto err2;
	}

	if(pipe2(queue,O_CLOEXEC))
	{
		perror("pipe2");
		goto err3;
	}

	if(prcprep(cfgfile,script,sfd,maxwait,trace))goto err4;

	if(!fg)if(daemon(0,0))goto err4;

	pfh[0]=efd;
	pfh[1]=queue[0];

	if(pthread_create(&prcthr,NULL,prcjob,pfh))
	{
		perror("pthread_create");
		goto err4;
	}

	if(pthread_create(&unxthr,NULL,unixjob,(void *)((long)efd)))
	{
		perror("pthread_create");
		goto err5;
	}

	nfh[0]=efd;
	nfh[1]=queue[1];
	nfh[2]=hack;

	if(pthread_create(&nfqthr,NULL,nfqjob,nfh))
	{
		perror("pthread_create");
		goto err5;
	}

	sigfillset(&s);
	sigdelset(&s,SIGHUP);
	sigdelset(&s,SIGINT);
	sigdelset(&s,SIGQUIT);
	sigdelset(&s,SIGTERM);
	sigdelset(&s,SIGCHLD);

	do
	{
		sigsuspend(&s);
	} while(!term);

err5:	if(write(efd,&dummy,sizeof(dummy))!=sizeof(dummy))
	{
		perror("write");
	}

	sigfillset(&s);
	sigdelset(&s,SIGCHLD);
	sigprocmask(SIG_SETMASK,&s,NULL);

	pthread_join(nfqthr,NULL);
	pthread_join(unxthr,NULL);
	pthread_join(prcthr,NULL);

err4:	close(queue[0]);
	close(queue[1]);
err3:	close(sfd);
err2:	close(efd);
err1:	exit(err);
}
