#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <net/if.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <linux/wireless.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <errno.h>
#include "nfqueue.h"
#include "processor.h"
#include "missing.h"

#define INIT	0
#define	ADDIF	1
#define	DELIF	2
#define ADDV4	3
#define DELV4	4
#define ADDV6	5
#define DELV6	6
#define WLAN	"wlan0"

typedef struct ifc
{
	struct ifc *next;
	int state;
	int v4total;
	int v6total;
	int v4alloc;
	int v6alloc;
	struct in_addr *v4;
	struct in6_addr *v6;
	char name[IFNAMSIZ+1];
} IFC;

static IFC *ifc;
static IFC wlan={NULL,DOWN,0,0,0,0,NULL,NULL,WLAN};
static int init=0;
static int queue=-1;
static int hack=0;
static unsigned int seq;
static char buffer[65536];

static int samsung_hack(char *ifname)
{
	int i;
	struct ifreq ifr;

	if((i=socket(AF_INET,SOCK_DGRAM,0))==-1)
	{
		perror("socket");
		return -1;
	}
	memset(&ifr,0,sizeof(ifr));
	strcpy(ifr.ifr_name,ifname);
	if(ioctl(i,SIOCGIFFLAGS,&ifr))
	{
		if(errno==ENODEV)
		{
			close(i);
			return 0;
		}
		perror("ioctl SIOCGIFFLAGS");
		close(i);
		return -1;
	}
	close(i);
	if(ifr.ifr_flags&IFF_UP)return 1;
	return 0;
}

static void dosend(IFC *d)
{
	int i;
	int total;
	struct iovec iov[7];

	if(!init)return;

	iov[0].iov_base=&total;
	iov[0].iov_len=sizeof(total);
	iov[1].iov_base=d->name;
	iov[1].iov_len=IFNAMSIZ+1;
	iov[2].iov_base=&d->state;
	iov[2].iov_len=sizeof(d->state);
	iov[3].iov_base=&d->v4total;
	iov[3].iov_len=sizeof(d->v4total);
	iov[4].iov_base=&d->v6total;
	iov[4].iov_len=sizeof(d->v6total);
	iov[5].iov_base=d->v4;
	iov[5].iov_len=d->v4total*sizeof(struct in_addr);
	iov[6].iov_base=d->v6;
	iov[6].iov_len=d->v6total*sizeof(struct in6_addr);

	for(total=0,i=1;i<7;i++)total+=iov[i].iov_len;

	if(writev(queue,iov,7)!=sizeof(total)+total)
		perror("writev");
}

static void initialize(void)
{
	IFC *d;
	int f=0;

	init=1;
	for(d=ifc;d;d=d->next)
	{
		if(!strcmp(d->name,WLAN))f=1;
		dosend(d);
	}
	if(hack&2)if(!f)dosend(&wlan);
}

static void handler(char *name,int mode,void *data)
{
	int i;
	IFC **e;
	IFC *d;
	void *tmp;
	struct in_addr *v4;
	struct in6_addr *v6;

	switch(mode)
	{
	case ADDIF:
		for(e=&ifc;*e;e=&(*e)->next)if(!strcmp(name,(*e)->name))break;
		if(!*e)
		{
			if(!(d=malloc(sizeof(IFC))))
			{
				perror("malloc");
				break;
			}
			memset(d,0,sizeof(IFC));
			strcpy(d->name,name);
			*e=d;
		}
		else d=*e;
		i=d->state;
		if((*((uint32_t *)data)&IFF_UP)==IFF_UP)d->state=UP;
		else d->state=DOWN;
		if(d->state!=i)dosend(d);
		break;

	case DELIF:
		for(e=&ifc;*e;e=&(*e)->next)if(!strcmp(name,(*e)->name))
		{
			if(hack&1)if(samsung_hack(name)==1)break;
			d=*e;
			*e=d->next;
			d->state=DOWN;
			d->v4total=0;
			d->v6total=0;
			dosend(d);
			if(d->v4alloc)free(d->v4);
			if(d->v6alloc)free(d->v6);
			free(d);
			break;
		}
		break;

	case ADDV4:
		for(d=ifc;d;d=d->next)if(!strcmp(name,d->name))
		{
			v4=(struct in_addr *)data;
			for(i=0;i<d->v4total;i++)
				if(d->v4[i].s_addr==v4->s_addr)break;
			if(i==d->v4total)
			{
				if(d->v4total==d->v4alloc)
				{
					if(!(tmp=realloc(d->v4,(d->v4alloc+4)*
						sizeof(struct in_addr))))
					{
						perror("realloc");
						break;
					}
					d->v4alloc+=4;
					d->v4=tmp;
				}
				d->v4[i].s_addr=v4->s_addr;
				d->v4total++;
				dosend(d);
			}
			break;
		}
		break;

	case DELV4:
		for(d=ifc;d;d=d->next)if(!strcmp(name,d->name))
		{
			v4=(struct in_addr *)data;
			for(i=0;i<d->v4total;i++)
				if(d->v4[i].s_addr==v4->s_addr)break;
			if(i!=d->v4total)
			{
				if(i+1!=d->v4total)
					memmove(&d->v4[i],&d->v4[i+1],
						(d->v4total-i-1)*
						sizeof(struct in_addr));
				d->v4total--;
				dosend(d);
			}
			break;
		}
		break;

	case ADDV6:
		for(d=ifc;d;d=d->next)if(!strcmp(name,d->name))
		{
			v6=(struct in6_addr *)data;
			for(i=0;i<d->v6total;i++)
				if(!memcmp(d->v6[i].s6_addr,v6->s6_addr,
					sizeof(struct in6_addr)))break;
			if(i==d->v6total)
			{
				if(d->v6total==d->v6alloc)
				{
					if(!(tmp=realloc(d->v6,(d->v6alloc+4)*
						sizeof(struct in6_addr))))
					{
						perror("realloc");
						break;
					}
					d->v6alloc+=4;
					d->v6=tmp;
				}
				memcpy(d->v6[i].s6_addr,v6->s6_addr,
					sizeof(struct in6_addr));
				d->v6total++;
				dosend(d);
			}
			break;
		}
		break;

	case DELV6:
		for(d=ifc;d;d=d->next)if(!strcmp(name,d->name))
		{
			v6=(struct in6_addr *)data;
			for(i=0;i<d->v6total;i++)
				if(!memcmp(d->v6[i].s6_addr,v6->s6_addr,
					sizeof(struct in6_addr)))break;
			if(i!=d->v6total)
			{
				if(i+1!=d->v6total)
					memmove(&d->v6[i],&d->v6[i+1],
						(d->v6total-i-1)*
						sizeof(struct in6_addr));
				d->v6total--;
				dosend(d);
			}
			break;
		}
		break;
	}
}

static int process(char *buffer,int len,unsigned int *seq)
{
	int rtl;
	uint32_t flags;
	struct nlmsghdr *nlh;
	struct ifaddrmsg *ifa;
	struct rtattr *rth;
	struct ifinfomsg *ifim;
	char name[IFNAMSIZ+1];

	nlh=(struct nlmsghdr *)buffer;

	while((NLMSG_OK(nlh,len)))
	{
		switch(nlh->nlmsg_type)
		{
		case NLMSG_DONE:
			*seq=nlh->nlmsg_seq;
			return 1;

		case RTM_NEWADDR:
               		ifa=(struct ifaddrmsg *)NLMSG_DATA(nlh);
               		rth=IFA_RTA(ifa);
               		rtl=IFA_PAYLOAD(nlh);
               		while(rtl&&RTA_OK(rth,rtl))
			{
				switch(rth->rta_type)
				{
				case IFA_LOCAL:
					if(ifa->ifa_family!=AF_INET)break;
					if_indextoname(ifa->ifa_index,name);
					handler(name,ADDV4,RTA_DATA(rth));
					break;

				case IFA_ADDRESS:
					if(ifa->ifa_family!=AF_INET6)break;
					if_indextoname(ifa->ifa_index,name);
					handler(name,ADDV6,RTA_DATA(rth));
					break;
				}
				rth=RTA_NEXT(rth,rtl);
			}
			break;

		case RTM_DELADDR:
               		ifa=(struct ifaddrmsg *)NLMSG_DATA(nlh);
               		rth=IFA_RTA(ifa);
               		rtl=IFA_PAYLOAD(nlh);
               		while(rtl&&RTA_OK(rth,rtl))
			{
				switch(rth->rta_type)
				{
				case IFA_LOCAL:
					if(ifa->ifa_family!=AF_INET)break;
					if_indextoname(ifa->ifa_index,name);
					handler(name,DELV4,RTA_DATA(rth));
					break;

				case IFA_ADDRESS:
					if(ifa->ifa_family!=AF_INET6)break;
					if_indextoname(ifa->ifa_index,name);
					handler(name,DELV6,RTA_DATA(rth));
					break;
				}
				rth=RTA_NEXT(rth,rtl);
			}
			break;

		case RTM_NEWLINK:
			ifim=(struct ifinfomsg *)NLMSG_DATA(nlh);
			if_indextoname(ifim->ifi_index,name);
			flags=ifim->ifi_flags;
			handler(name,ADDIF,&flags);
			break;

		case RTM_DELLINK:
			ifim=(struct ifinfomsg *)NLMSG_DATA(nlh);
			if_indextoname(ifim->ifi_index,name);
			flags=ifim->ifi_flags;
			handler(name,DELIF,&flags);
			break;
		}

		nlh=NLMSG_NEXT(nlh,len);
	}

	return 0;
}

static int getaddrlist(int fd,int mode)
{
	struct
	{
		struct nlmsghdr n;
		struct ifaddrmsg r;
		char data[16];
	} req;
	struct rtattr *rta;

	memset(&req,0,sizeof(req));
	req.n.nlmsg_len=NLMSG_LENGTH(sizeof(struct ifaddrmsg));
	req.n.nlmsg_flags=NLM_F_REQUEST|NLM_F_ROOT;
	req.n.nlmsg_type=RTM_GETADDR;
	req.n.nlmsg_seq=++seq;
	req.r.ifa_family=mode?AF_INET6:AF_INET;
	rta=(struct rtattr *)(((char *)&req)+NLMSG_ALIGN(req.n.nlmsg_len));
	rta->rta_len=RTA_LENGTH(4);
	if(send(fd,&req,req.n.nlmsg_len,0)<0)return -1;
	return 0;
}

static int getlinklist(int fd)
{
	struct
	{
		struct nlmsghdr n;
		struct rtgenmsg r;
	} req;

	memset(&req,0,sizeof(req));
	req.n.nlmsg_len=NLMSG_LENGTH(sizeof(struct rtgenmsg));
	req.n.nlmsg_flags=NLM_F_REQUEST|NLM_F_DUMP;
	req.n.nlmsg_type=RTM_GETLINK;
	req.n.nlmsg_seq=++seq;
	req.r.rtgen_family=AF_PACKET;;
	if(send(fd,&req,req.n.nlmsg_len,0)<0)return -1;
	return 0;
}

void *nfqjob(void *data)
{
	int len;
	int state;
	unsigned int ans;
	struct pollfd p[2];
	struct sockaddr_nl addr;
	sigset_t s;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	p[0].fd=((int *)data)[0];
	queue=((int *)data)[1];
	hack=((int *)data)[2];

	seq=time(NULL);
	state=0;
	ifc=NULL;

	if((p[1].fd=socket(PF_NETLINK,SOCK_RAW|SOCK_CLOEXEC,NETLINK_ROUTE))==-1)
	{
		perror("couldn't open NETLINK_ROUTE socket");
		goto fail;
	}

	len=sizeof(buffer);
	if(setsockopt(p[1].fd,SOL_SOCKET,SO_SNDBUF,&len,sizeof(len)))
	{
		perror("setsockopt SO_SNDBUF\n");
		goto fail;
	}
	if(setsockopt(p[1].fd,SOL_SOCKET,SO_RCVBUF,&len,sizeof(len)))
	{
		perror("setsockopt SO_RCVBUF\n");
		goto fail;
	}

	memset(&addr,0,sizeof(addr));
	addr.nl_family=AF_NETLINK;
	addr.nl_groups=RTMGRP_IPV4_IFADDR|RTMGRP_IPV6_IFADDR|RTNLGRP_LINK;

	if(bind(p[1].fd,(struct sockaddr *)&addr,sizeof(addr))==-1)
	{
		perror("couldn't bind");
		goto fail;
	}

	if(getlinklist(p[1].fd))
	{
		perror("getlinklist send");
		goto fail;
	}

	p[0].events=POLLIN;
	p[1].events=POLLIN;

	while(1)
	{
		if((len=poll(p,2,-1))==-1)continue;

		if(!len)continue;

		if(p[0].revents&POLLIN)goto out;

		if(p[1].revents&POLLIN)
			if((len=recv(p[1].fd,buffer,sizeof(buffer),0))>0)
				if(process(buffer,len,&ans))switch(state)
		{
		case 0:
		case 1: if(ans==seq)if(getaddrlist(p[1].fd,state++))
			{
				perror("getaddrlist send");
				goto fail;
			}
			break;

		case 2:	if(ans==seq)
			{
				initialize();
				state++;
			}
			break;
		}
	}

fail:	kill(getpid(),SIGTERM);

out:	if(p[1].fd!=-1)close(p[1].fd);

	pthread_exit(NULL);
	return NULL;
}
