#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <net/if.h>
#include "processor.h"
#include "parse.h"

#define MAX_STATE	31
#define MAX_STATES	16

typedef struct state
{
	struct state *next;
	int type;
	int match;
	char state[MAX_STATE+1];
	char dev[IFNAMSIZ+1];
} STATE;

typedef struct label
{
	struct label *next;
	int start;
	int stop;
	int total;
	char label[MAX_LABEL+1];
	char states[MAX_STATES][MAX_STATE+1];
	OPSTATE *ptr[MAX_STATES];
} LABEL;

static int sorter(const void *s1,const void *s2)
{
	return strcmp((char *)s1,(char *)s2);
}

static int dolabel(char *in,LABEL *data)
{
	int i;
	int j;
	char *label;
	char *start;
	char *stop;
	char *state;
	char *ptr;

	label=strtok(in," \t:");
	start=strtok(NULL," \t\r\n");
	stop=strtok(NULL," \t\r\n");

	if(!label||!*label||!start||!*start||!stop||!*stop)return -1;

	if(strlen(label)>MAX_LABEL)return -1;
	strcpy(data->label,label);

	if(strlen(start)>2)return -1;
	for(ptr=start;*ptr;ptr++)if(!(*ptr>='0'&&*ptr<='9'))return -1;
	data->start=atoi(start);

	if(strlen(stop)>2)return -1;
	for(ptr=stop;*ptr;ptr++)if(!(*ptr>='0'&&*ptr<='9'))return -1;
	data->stop=atoi(stop);

	for(i=0;;i++)
	{
		if(!(state=strtok(NULL," \t\r\n")))break;
		if(!*state||strlen(state)>MAX_STATE)return -1;
		if(i==MAX_LABEL)return -1;

		for(j=0;j<i;j++)if(!strcmp(data->states[i],state))return -1;

		strcpy(data->states[i],state);
	}
	if(!i)return -1;

	data->total=i;
	data->next=NULL;

	qsort(data->states,data->total,MAX_STATE+1,sorter);

	for(i=1;i<data->total;i++)if(!strcmp(data->states[i-1],data->states[i]))
		return -1;

	return 0;
}

static int dostate(char *in,STATE *data)
{
	int type;
	int match;
	char *state;
	char *dev;
	char *mode;
	char *oper;

	state=strtok(in," \t=");
	dev=strtok(NULL," \t\r\n");
	mode=strtok(NULL," \t\r\n");

	if(!state||!*state||!dev||!*dev||!mode||!*mode)return -1;

	if(strlen(state)>MAX_STATE)return -1;
	if(strlen(dev)>IFNAMSIZ)return -1;

	if(!strcmp(mode,"down"))
	{
		type=DOWN;
		match=NOIP;

		if(strtok(NULL," \t\r\n"))return -1;
		goto save;
	}

	if(!strcmp(mode,"up"))
	{
		type=UP;

		oper=strtok(NULL," \t\r\n");

		if(!oper)
		{
			match=NOIP;
			goto save;
		}

		if(!strcmp(oper,"ipv4"))match=IPV4;
		else if(!strcmp(oper,"ipv6"))match=IPV6;
		else if(!strcmp(oper,"ip"))match=IP;
		else return -1;

		if(strtok(NULL," \t\r\n"))return -1;
		goto save;
	}

	if(!strcmp(mode,"change"))
	{
		type=CHANGE;

		if(!(oper=strtok(NULL," \t\r\n")))return -1;

		if(!strcmp(oper,"ipv4"))match=IPV4;
		else if(!strcmp(oper,"ipv6"))match=IPV6;
		else if(!strcmp(oper,"ip"))match=IP;
		else return -1;

		if(strtok(NULL," \t\r\n"))return -1;
		goto save;
	}

	return -1;

save:	data->type=type;
	data->match=match;
	strcpy(data->state,state);
	strcpy(data->dev,dev);
	data->next=NULL;

	return 0;

}

int parse(char *fn,OPDEV **oda,OPCOND **oca)
{
	int i;
	int err=-1;
	int line;
	char *ptr;
	FILE *fp;
	STATE *se;
	LABEL *le;
	STATE *slist=NULL;
	LABEL *llist=NULL;
	STATE **sa=&slist;
	LABEL **la=&llist;
	OPDEV *odb=*oda;
	OPDEV *ode;
	OPSTATE *ose;
	OPCOND *oce;
	STATE state;
	LABEL label;
	char bfr[1024];

	if(!(fp=fopen(fn,"r")))
	{
		fprintf(stderr,"can't open %s\n",fn);
		return -1;
	}

	for(ptr=bfr,line=1;fgets(bfr,sizeof(bfr),fp);line++,ptr=bfr)
	{
		while(*ptr==' '||*ptr=='\t')ptr++;
		if(!*ptr||*ptr=='\r'||*ptr=='\n'||*ptr=='#')continue;

		for(ptr++;*ptr;ptr++)if(*ptr=='=')break;
		else if(*ptr==':')break;

		switch(*ptr)
		{
		case 0:	fprintf(stderr,"syntax error line %d\n",line);
			goto fail;

		case '=':
			if(dostate(bfr,&state))
			{
				fprintf(stderr,"syntax error line %d\n",line);
				goto fail;
			}

			for(se=slist;se;se=se->next)
				if(!strcmp(se->state,state.state))
			{
				fprintf(stderr,"duplicate state line %d\n",
					line);
				goto fail;
			}
			else if(!strcmp(se->dev,state.dev)&&
				se->type==state.type&&se->match==state.match)
			{
				fprintf(stderr,"duplicate state line %d\n",
					line);
				goto fail;
			}

			if(!(se=malloc(sizeof(STATE))))
			{
				perror("malloc");
				goto fail;
			}

			*se=state;
			*sa=se;
			sa=&se->next;
			break;

		case ':':
			if(dolabel(bfr,&label))
			{
				fprintf(stderr,"syntax error line %d\n",line);
				goto fail;
			}

			for(le=llist;le;le=le->next)
				if(!strcmp(le->label,label.label))
			{
				fprintf(stderr,"duplicate label line %d\n",
					line);
				goto fail;
			}
			else if(le->start==label.start&&le->stop==label.stop&&
				le->total==label.total)
			{
				for(i=0;i<label.total;i++)
				    if(strcmp(le->states[i],label.states[i]))
					break;

				if(i==label.total)
				{
					fprintf(stderr,"duplicate label data "
						"line %d\n",line);
					goto fail;
				}
			}

			for(i=0;i<label.total;i++)
			{
				for(se=slist;se;se=se->next)
					if(!strcmp(se->state,label.states[i]))
						break;

				if(!se)
				{
					fprintf(stderr,"state %s not found on "
						"line %d\n",label.states[i],
						line);
					goto fail;
				}
			}

			if(!(le=malloc(sizeof(LABEL))))
			{
				perror("malloc");
				goto fail;
			}

			*le=label;
			*la=le;
			la=&le->next;
			break;
		}
	}

	for(se=slist;se;se=se->next)
	{
		for(ode=odb;ode;ode=ode->next)if(!strcmp(ode->dev,se->dev))
			break;

		if(!ode)
		{
			if(!(ode=malloc(sizeof(OPDEV)+strlen(se->dev)+1)))
			{
				perror("malloc");
				goto fail;
			}

			memset(ode,0,sizeof(OPDEV));
			strcpy(ode->dev,se->dev);
			*oda=ode;
			oda=&ode->next;
		}

		if(!(ose=malloc(sizeof(OPSTATE)+strlen(se->state)+1)))
		{
			perror("malloc");
			goto fail;
		}

		ose->next=ode->states[se->type];
		ose->type=se->type;
		ose->match=se->match;
		ose->valid=0;
		strcpy(ose->name,se->state);
		ode->states[se->type]=ose;

		for(le=llist;le;le=le->next)for(i=0;i<le->total;i++)
			if(!strcmp(se->state,le->states[i]))
		{
			le->ptr[i]=ose;
			break;
		}
	}

	for(le=llist;le;le=le->next)
	{
		if(!(oce=malloc(sizeof(OPCOND)+le->total*sizeof(OPSTATE *))))
		{
			perror("malloc");
			goto fail;
		}

		oce->next=NULL;
		oce->start=le->start*1000;
		oce->stop=le->stop*1000;
		oce->total=le->total;
		oce->currstate=STOPPED;
		oce->nextstate=STOPPED;
		oce->delay=0;
		oce->event=0;
		strcpy(oce->label,le->label);
		memcpy(oce->states,le->ptr,le->total*sizeof(OPSTATE *));

		*oca=oce;
		oca=&oce->next;
	}

	err=0;

fail:	fclose(fp);

	while(slist)
	{
		se=slist;
		slist=se->next;
		free(se);
	}

	while(llist)
	{
		le=llist;
		llist=le->next;
		free(le);
	}

	return err;
}
